/*
  rup_bro_sor.c  --  rup  --  host status broadcast query

  Copyright (C) Kazuro Furukawa, 1988-2001
  Kazuro Furukawa, KEK, Japan,  kazuro.furukawa@kek.jp

  developed as an example of broadcast sun-rpc call
  based on rstat.c from surpc 4.0 and my rusers.c

  nov.14.1992.
  version 1	query one host (rup_int.c or rup.c)
  version 2	query with broadcast (rup_bro.c or rup2.c)
  version 3	query with broadcast and no duplicate (rup_bro_sel.c or rup3.c)
  		(home made list version)
  nov.21.1992.
  version 4	same as version 3 but use tsearch (rup4.c)
  version 5	same as version 4 but process options as sun's (rup5.c)
  nov.24.1992.  remove ANSI statements
  mar.6.1993.	search time option (-s)
  mar.31.1993.	output format of load average
  feb.8.2001.   using alarm to timeout 
  oct.29.2001.  fix alarm bug
  nov.13.2001.  fix alarm bug, add -n option

  rup queries host status with sun-rpc rstatvers_time service
  it calls the service using broadcast and display response one by one
  if an option is specified, it sorts the results before displaying
  if hostnames are specified, it displays their responses

  it uses tsearch routines to avoid duplicate information
  it uses qsort routine to sort results

  rup			broadcast query
  rup [-{h|l|t}r]	sort responses
		-t: sort by uptime, -l: by load, -h: by hostname, -r: reverse
  rup [-{h|l|t}r] [-n]	show ip address instead of host name with sort
  rup [-sNN]		search NN seconds
  rup [-v]		verbose mode
  rup [hosts]		query hosts

*/

char    rup_what[] =
"@(#)rup.c v5.5.3 rstatvers_time rpc query, K.Furukawa,Kek, Nov.1992-Nov.2001";

char    * rup_usage[] = {
    "usage: rup [-{t|l|h}r] [-n] [-sNN] [hosts ...]\n",
    "  -t: sort by uptime,  -l: by load, -h: by hostname,  -r: reverse\n",
    "  -n: numerical address (only with sort options),     -v: verbose\n",
    "  -s NN: search time NN seconds (default is 60)\n",
    0
    };

/* #define DEBUG */

#include <stdio.h>
#include <rpc/rpc.h>
#include <rpc/pmap_clnt.h>
#include "rstat.h"			/* to be generated from rstat.x */
#include <time.h>			/* tm structure, no need on sun4.0.3 */
#include <sys/socket.h>			/* socket structures */
#include <netdb.h>			/* getxxx routines */
#include <arpa/inet.h>                  /* inet_xxx routines */
#include <search.h>			/* tsearch routines */
#include <unistd.h>			/* alarm */
#include <signal.h>			/* alarm */
#include <setjmp.h>			/* longjmp */

#define TIMEOUT 60			/* default search time */
#define BUFFERSIZE 1600			/* buffer size of result on ethernet */

struct statlocal  {			/* stat info element */
    struct in_addr addr;		/* server internet address */
    time_t uptime;			/* uptime info */
    int avenrun[3];			/* load average */
    char hostname[16];			/* host name */
};
typedef struct statlocal StatLocal;

time_t starttime;			/* program start time */

StatLocal *root_stat;			/* pointer to root of tree */
StatLocal *temp_stat;			/* temporary area pointer */
StatLocal **list_stat;			/* list of stat info */

int timeout = TIMEOUT;			/* search time */
int counter = 0;			/* number of valid responses */
int reverse = 1;			/* sort forward/reverse */
int dotaddr = 0;			/* reverse name resolution */
int verbose = 0;			/* some warning messages */

sigjmp_buf jmpbuf;			/* longjmp buffer */

void (*display)();			/* display routine */
void (*sort)();				/* sort routine */
int (*compare)();			/* compare routine */

void alarm_handler( int signo ) {
    siglongjmp(jmpbuf,1);
}

void main(argc, argv)
int argc;
char **argv;
{
    static enum clnt_stat clnt_stat;
    static char buffer[BUFFERSIZE];		/* buffer to receive result */
    bool_t process();				/* process each result */
    void real_display();			/* display routine */
    void dummy();				/* dummy routine */
    void option();				/* option handler */
    void fill_stat();				/* convert response to local */

    display = real_display;			/* display routine */
    sort = dummy;				/* dummy by default */

    if ( !(temp_stat =				/* first temporary area */
	 (StatLocal *)malloc( sizeof (StatLocal) ) ) )  {
	perror("malloc");			/* test to allocate */
	exit(1);
    }
    bzero(&temp_stat->addr, sizeof (struct in_addr));

    option(&argc, &argv);			/* handle options */

    if ( argc <= 1 )  {				/* if no argument specified */

	if (sort != dummy)			/* sorting option specified */
	    fprintf(stderr, "collecting responses...");

	root_stat = (void *)NULL;
	time(&starttime);			/* get start time */

	clnt_stat = 				/* broadcast call */
 	    clnt_broadcast(RSTATPROG, RSTATVERS_TIME, RSTATPROC_STATS,
		       xdr_void, 0, xdr_statstime, buffer, process);
	if (clnt_stat != RPC_SUCCESS && clnt_stat != RPC_TIMEDOUT) {
	    clnt_perrno(clnt_stat);		/* error other than time out */
	    fprintf(stderr, "\n");
	    /* exit(-1); */
	}

	sort();					/* sort and display */

    }  else  {					/* arguments specified */
	static int i;
	signal( SIGALRM, alarm_handler );	/* install alarm handler */
	for ( i=1; i<argc; i++ )  {
	    static int rpcstat;
	    static StatLocal *thisstat;
	    /* timeout capability in clnt_create, etc did not work */
	jumptarget:
	    if (sigsetjmp(jmpbuf,0) != 0) {
		/* comes here from alarm handler */
		if ( verbose ) printf("Time-out\n");
		i++;				/* next host */
		if (i>=argc) break;		/* escape from for loop */
		signal( SIGALRM, alarm_handler ); /* install alarm handler */
		goto jumptarget;
	    }
	    alarm( timeout );
	    if ( rpcstat = callrpc( argv[i], RSTATPROG, RSTATVERS_TIME,
	         RSTATPROC_STATS, xdr_void, 0, xdr_statstime,
	         buffer) != 0 )  {
		clnt_perrno(rpcstat);
		fprintf(stderr, "\n");
		/* exit(-1); */	
		if ( verbose ) printf("Error\n");
		alarm( 0 );
		continue;			/* ignore result */
	    }
	    alarm( 0 );
	    thisstat = temp_stat;
	    fill_stat( thisstat, buffer, NULL ); /* convert response */
	    strncpy( thisstat->hostname, argv[i], 15 );
	    real_display( thisstat );		/* display response */
	}
    }
    exit(0);
}

/* dummy  --  do nothing */

void dummy()
{
    ;
}

/* option  --  handle command line options */

void option( argc, argv )
int *argc;
char ***argv;
{
    int c;
    extern int optind, opterr;
    extern char *optarg;
    int uptime_compare();     		/* compare routines */
    int load_compare();
    int host_compare();
    void sort_list();
    char *argtmp;
    opterr = 0;				/* disable informational message */

    while ((c = getopt (*argc, *argv, "tlhrvVns:")) != EOF)  {
        switch (c) {
	case 't':
	    display = dummy;			/* no display during process */
	    sort = sort_list;			/* invoke sort routine */
	    compare = uptime_compare;		/* sort by uptime */
	    break;
	case 'l':
	    display = dummy;			/* no display during process */
	    sort = sort_list;			/* invoke sort routine */
	    compare = load_compare;		/* sort by load average */
	    break;
	case 'h':
	    display = dummy;			/* no display during process */
	    sort = sort_list;			/* invoke sort routine */
	    compare = host_compare;		/* sort by host name */
	    break;
	case 'r':
	    reverse = -1;			/* reverse sort */
	    break;
        case 's':
            timeout = atoi(optarg);		/* search time */
            if (!timeout)  {
                fprintf(stderr, "illegal search time, default %d used\n",
                        TIMEOUT);
                timeout = TIMEOUT;
            }
            break;
        case 'v':
            verbose++;
            break;
        case 'V':
            fputs(rup_what, stderr);
            fputs("\n", stderr);
            break;
        case 'n':
            dotaddr++;
            break;
	case '?':
	    fputs("illegal option\n", stderr);
            {
                int i = 0;
                while(rup_usage[i])
                    fputs(rup_usage[i++], stderr);
            }
	    exit(-1);
	}
    }
    *argc -= optind - 1;			/* replace argc/argv */
    argtmp = (*argv)[0];
    *argv = &((*argv)[optind-1]);
    (*argv)[0] = argtmp;
}

/* process  --  process each response, control display */

bool_t process( statinfo, addr )
statstime * statinfo;			/* RSTATPROC_STATS result */
struct sockaddr_in *addr;		/* server address */
{
    {
	StatLocal ** thisstat;
	int addr_compare();
	void fill_stat();

	temp_stat->addr = addr->sin_addr;	/* get server address */

#ifdef DEBUG
	printf("temp_stat: %x, temp_stat->addr: %x\n",
	       temp_stat, temp_stat->addr.s_addr);
#endif

	if( !(thisstat = 
	   (StatLocal **) tsearch (temp_stat, &root_stat, addr_compare) ) )  {
	    perror("tsearch");			/* insert into btree */
	    exit(1);
	}

#ifdef DEBUG
	printf("*thisstat: %x, *thisstat->addr: %x\n",
	       *thisstat, (*thisstat)->addr.s_addr);
#endif

	if (*thisstat == temp_stat)  {		/* new host info */
	    fill_stat(*thisstat, statinfo, addr); /* fill out info */
	    display( *thisstat );		/* display if not duplicate */
	    counter++;				/* add one counter */
	}
    }

    {
	if ((time((time_t *)NULL) - starttime) > timeout )
	    return TRUE;			/* time out */
	else
	    return FALSE;			/* continue to query */
    }
}

/* check_host  --  convert internet address into host name */

void check_host( statinfo )
StatLocal * statinfo;			/* info to display */
{
    if (!((int)statinfo->hostname[0])||dotaddr)  {
	struct	hostent	*hp;			/* hostname entry */

	if ( dotaddr )
	    hp = NULL;
	else
	    hp = gethostbyaddr( (char *) &statinfo->addr, 
			       sizeof( struct in_addr ), AF_INET );
						/* get server information */
	if ( hp != NULL )  {
	    strncpy(statinfo->hostname,hp->h_name,15); /* host name */
	}  else  {
	    strncpy(statinfo->hostname,inet_ntoa(statinfo->addr),15);
						/* dot format */
  	}
    }
}

/* real_display  --  display info */

void real_display( statinfo )
StatLocal * statinfo;			/* info to display */
{
    check_host( statinfo );			/* check hostname */
    printf("%15.15s ", statinfo->hostname);	/* print hostname */

    {
	struct tm   *tmp_time;
	char         days_buf[16];
	char         hours_buf[16];

	tmp_time = gmtime(&statinfo->uptime);

	if (tmp_time->tm_yday != 0)
	    sprintf(days_buf, "%4d day%s ", tmp_time->tm_yday,
		    (tmp_time->tm_yday > 1) ? "s," : ", ");
	else
	    strcpy(days_buf,"           ");

	if (tmp_time->tm_hour != 0)  {
	    if (tmp_time->tm_min !=0)
		sprintf(hours_buf, "%2d:%02d,   ",
			tmp_time->tm_hour, tmp_time->tm_min);
	    else
		sprintf(hours_buf, "%2d hour%s", tmp_time->tm_hour,
			(tmp_time->tm_min == 1) ? ", " : "s,");
	}  else  {
	    sprintf(hours_buf, "%2d min%s", tmp_time->tm_min,
		    (tmp_time->tm_min == 1) ? ",  " : "s, ");
	}

	printf(" up%s%s load average:%5.2f,%5.2f,%5.2f\n",
	       days_buf,
	       hours_buf,
	       (double)statinfo->avenrun[0]/FSCALE,
	       (double)statinfo->avenrun[1]/FSCALE,
	       (double)statinfo->avenrun[2]/FSCALE);
    }
}

/* fill_stat  --  fill info into element and allocate next temp. element */

void fill_stat(stat, statinfo, addr)
StatLocal *stat;			/* insertion point */
statstime * statinfo;			/* RSTATPROC_STATS result */
struct sockaddr_in *addr;		/* server address */
{
    if ( addr != NULL )
	stat->addr = addr->sin_addr;
						/* get server address */

    statinfo->curtime.tv_sec -= statinfo->boottime.tv_sec;

    stat->uptime = statinfo->curtime.tv_sec;	/* get server uptime */

    stat->avenrun[0] = statinfo->avenrun[0];	/* get load averages */
    stat->avenrun[1] = statinfo->avenrun[1];
    stat->avenrun[2] = statinfo->avenrun[2];

    stat->hostname[0] = '\0';

    if ( !(temp_stat = 
	   (StatLocal *)malloc( sizeof (StatLocal) ) ) )  {
	perror("malloc");			/* allocate next element */
	exit(1);
    }
}

/* addr_compare  --  compare address, for tsearch routine */

int addr_compare(stat1,stat2)
StatLocal *stat1, *stat2;		/* element pointers */
{
#ifdef DEBUG
    printf("1: %x, 2: %x\n",stat1->addr.s_addr,stat2->addr.s_addr);
#endif

    /* because of overflow, we can not return difference */

    if (stat1->addr.s_addr == stat2->addr.s_addr)
	return 0;				/* duplicate address */
    if (stat1->addr.s_addr > stat2->addr.s_addr)  return 1;
    else return -1;
}

/* uptime_compare  --  compare by uptime, for sort routine */

int uptime_compare(stat1,stat2)
StatLocal **stat1, **stat2;
{
    if ((*stat1)->uptime == (*stat2)->uptime)  return 0;
    if ((*stat1)->uptime < (*stat2)->uptime)  return reverse;
    else return -reverse;
}

/* load_compare  --  compare by load average, for sort routine */

int load_compare(stat1,stat2)
StatLocal **stat1, **stat2;
{
    if ((*stat1)->avenrun[0] == (*stat2)->avenrun[0])  return 0;
    if ((*stat1)->avenrun[0] > (*stat2)->avenrun[0])  return reverse;
    else return -reverse;
}

/* host_compare  --  compare by host name, for sort routine */

int host_compare(stat1,stat2)
StatLocal **stat1, **stat2;
{
    check_host( *stat1 );			/* convert address to name */
    check_host( *stat2 );
    return (reverse * strncmp( (*stat1)->hostname, (*stat2)->hostname, 15) );
}

/* sort_list  --  sort stat info */

void sort_list()
{
    StatLocal **list;			/* array of pointer to info */
    StatLocal **make_list();
    int i;

    fputs("\n", stderr);

    list = make_list();				/* make list from btree */

    qsort( list, counter, sizeof (void *), compare); /* sort list */

    for (i=0; i<counter; i++)  {		/* display each element */
#ifdef DEBUG
	printf("*list->addr: %x\n",(*list)->addr.s_addr);
#endif
	real_display(*list++);
    }
}

/* make_list  --  make list of all responses */

StatLocal **make_list()			/* array of pointer to info */
{
    int fill_list();
    StatLocal **tmplist_stat;

    if ( !(list_stat = 
	   (StatLocal **)malloc( counter * sizeof (void *) ) ) )  {
	perror("malloc");			/* allocate next element */
	exit(1);
    }
    tmplist_stat = list_stat;

    twalk( root_stat, fill_list );		/* fill each info */

    return tmplist_stat;
}

/* fill_list  --  fill each list with a response */

int fill_list(stat,  visit,  level)
StatLocal **stat;
VISIT visit;
int level;
{
    if (visit == preorder || visit == leaf)  {
#ifdef DEBUG
	printf("*stat->addr: %x\n",(*stat)->addr.s_addr);
#endif
	*list_stat++ = *stat;			/* insert into list */
    }
    return 1;
}
